#import "@preview/polylux:0.3.1": *
#import "@preview/codelst:2.0.1": sourcecode
#import "@preview/cades:0.3.0": qr-code
#import "@preview/bob-draw:0.1.0": render
#import "@preview/oxifmt:0.2.1": strfmt

#set page(paper: "presentation-16-9")
#set text(size: 25pt)

#let centred-slide(body) = {
  polylux-slide(align(horizon + center, body))
}

#let list-slide(title, body) = {
  polylux-slide([
    #align(horizon + center, [= #title])
    #body
  ])
}

#let r(new, base) = {
  strfmt("{:.1}%",100-100*new / base)
}

#let v20 = v(20pt)
#let v30 = v(30pt)
#let v40 = v(40pt)
#let v60 = v(60pt)

#let s(body) = text(size:20pt, body)

#centred-slide[
  = Size of Rust binaries

  Link to this presentation:
    
  #qr-code("https://vi-server.org/pub/rustsize/presentation.pdf", width: 6cm)
]

#centred-slide[
  = Link to the code samples
    
  #v60

  #qr-code("https://vi-server.org/pub/rustsize/code.zip", width: 8cm)
]

#centred-slide[
  = Part 1: bloated programs
]

#list-slide("Typical bloated program")[
  #v40
#render(```
+-+-------+-------------------------------+---+--------------+
| |stdlib | dependencies                  | ! |  debugging   |
+-+-------+-------------------------------+---+--------------+
 ^                                         ^   "(and other"
 |                                your ___/     error
  \__ something                   code            "handling )"
      strictly necessary
      for the system
  
```,width: 80%)

#v20

`size = C + k * project_size`
]


#centred-slide[
  = Tradeoffs
  #v40
#render(```
                     fast execution
  _________________/
                   \
  fast               small
  compilation        size



  ----------------------------------
  debuggability           small size

```,width: 60%)

]

#centred-slide[
  = Reasons why big

  #v40

  `size = C + k * project_size`

  #v40

  - No dynamic linking
  - Monomorphisation
  - Debugging features
  - Stdlib algorithms
]



#centred-slide[
  = Reasons why big: No dynamic linking

  #v40

Typical Rust-generated executable or cdylib is mostly *self-sufficient* and reaches for *OS-supplied libraries* only when really needed.
]

#centred-slide[
  = Reasons why big: Monomorphisation

  #v60

  Static dispatch is popular and leads to better performance, but may bloat up the code.

  *Intruducting* otherwise unneeded *`dyn`* can help with battling the bloat. There are crates to automate that in some cases.
]

#centred-slide[
  = Reasons why big: Debugging features

  #v40

  Rust executables typically contain *helpful messages* to be displayed in case of panic.
]


#centred-slide[
  = Reasons why big: stdlib algorithms

  #v20

  Typically algorithms in stdlib (e.g. sorting) are *aimed at performance*, not compactness.
]


#list-slide([Reasons for pursuing minsized])[
  #v(30pt)
  - Getting ready for the *embedded* with access to hold tools
  - *Security* review of dependencies
    - Less bytes - *less code*
    - *Gamified* - nice to see the number shrink as you go though the deps.
  - *Aesthetical* resons: bloat=bad.
  - *Educational* exercise.
]


#centred-slide[
  = Part 2A: Shrinking challenge

  = (normal measures)
]


#list-slide("Sample project")[
  #v30
  Let's make a *TCP forwarder* as a sample project.

  - userspace
  - async
  - single threaded
  - CLI arguments with a help message

  #v20

  #s[Note: used `rustc 1.84.0-nightly (bc5cf994d 2024-11-05)`]
]


#centred-slide[
  = Basic version source code (part 1)
#sourcecode[```rust
use clap::Parser;    use std::net::SocketAddr;
use tokio::net::{TcpListener, TcpStream};
#[derive(Parser)]    struct Args {
    listen_addr: SocketAddr,
    connect_addr: SocketAddr,
}
#[tokio::main(flavor="current_thread")]
async fn main() -> anyhow::Result<()> {
    ... 
    Ok(()) 
}
```]
]

#centred-slide[
  = Basic version source code (part 2)
#sourcecode[```rust
    let Args { listen_addr, connect_addr } = Args::parse();
    let l = TcpListener::bind(listen_addr).await?;
    while let Ok((mut s, ..)) = l.accept().await {
        tokio::spawn(async move {
            let mut c = TcpStream::connect(
                                  connect_addr).await?;
            tokio::io::copy_bidirectional(
                                &mut s, &mut c).await?;
            Ok::<(),anyhow::Error>(())
        });
    }
```]
]

#centred-slide[
  = Basic version's metrics

  #v30

  `cargo build --release --target=x86_64-unknown-linux-gnu` + `strip`

  #table(
    columns: 2,
    [*target*], [*size*],
    [amd64 linux],[1308k],
    [i686 musl], [1304k],
    [arm64 android], [1180k],
    [windows64], [2212k],
    [arm64 mac], [1172k],
  )
]


#centred-slide[
  = Other languages (for a comparison)

  #v30

   #table(
    columns: 3,
    [*target*],     [*Go size*], [*Haskell size*],
    [amd64 linux],   [2136k],  [2976k],  
    [i686 linux],    [1968k],  [],  
    [arm64 android], [2372k],  [],  
    [windows64],     [2216k],  [],  
    [arm64 mac],     [2168k],  [],  
  )

  #v20

  With *tinygo* and appropriate `Netdev` I expect 100-200k to be reachable.
]


#list-slide(["Free beer" optimisations])[

  #v60

  - LTO
  - codegen units = 1
  - opt-level = "s"
]

#centred-slide[
  = "Free beer" reduction

  #v60

  #table(
    columns: 4,
    [*target*], [*earlier size*], [*new base size*], [*reduction*],
    [amd64 linux],   [1308k],  [848k ], [#r(848, 1308)],
    [i686 musl],     [1304k],  [908k ], [#r(908 ,1304)],
    [arm64 android], [1180k],  [776k ], [#r(776 ,1180)],
    [windows64],     [2212k],  [772k ], [#r(772 ,2212)],
    [arm64 mac],     [1172k],  [764k ], [#r(764 ,1172)],
  )
]

#list-slide([Deeper compilation mode changes])[
  Lossy.

   ```
   cargo build --profile=myrelease \
       -Z build-std=panic_abort,std \
       -Z build-std-features=panic_immediate_abort \
       --target=x86_64-unknown-linux-gnu
   ```

  - no unwinding
  - build_std
  - panic_immediate_abort
]


#list-slide([Other notable unstable settings])[
  #v60

  - `-Z build-std-features=optimize_for_size`
  - `RUSTFLAGS="-Zlocation-detail=none"`
  - `RUSTFLAGS="-Zfmt-debug=none"`
  - `-Zvirtual-function-elimination`
  - `-Zno-jump-tables`
]


#centred-slide[
  = Lean build results

  #v40

  #table(
    columns: 4,
    [*target*], [*base size*], [*lean size*], [*reduction*],
    [amd64 linux],   [848k],  [400k ], [#r(400,848)],
    [i686 musl],     [908k],  [484k ], [#r(484,908)],
    [arm64 android], [776k],  [416k ], [#r(416,776)],
    [windows64],     [772k],  [448k ], [#r(448,772)],
    [arm64 mac],     [764k],  [360k ], [#r(360,764)],
  )
]

#centred-slide[
  = Offending crates

  #v40


  #block[

  #align(left)[`cargo bloat --crates`]

  #v20

  ```
 File  .text     Size Crate
 4.7%  45.6% 285.5KiB std
 3.4%  33.5% 209.5KiB clap_builder
 1.2%  11.8%  73.7KiB tokio
 0.4%   3.8%  24.1KiB base (our crate)
  ```

  ]
]

#centred-slide[
  = Easy measure: limit `clap`'s features

  #v30

  `cargo add clap --no-default-features -F help,derive,std`
  
  #v20

   #table(
    columns: 5,
    [*target*], [*base size*], [*base reduction*], [*lean size*], [*lean reduction*],
    [amd64 linux],   [768k],  [#r(768,848)], [336k ], [#r(336,400)],
    [i686 musl],     [840k],  [#r(840,908)], [428k ], [#r(428,484)],
    [arm64 android], [708k],  [#r(708,776)], [360k ], [#r(360,416)],
    [windows64],     [680k],  [#r(680,772)], [368k ], [#r(368,448)],
    [arm64 mac],     [700k],  [#r(700,764)], [312k ], [#r(312,360)],
  )
]

#centred-slide[
  = Replacing "clap" with "argh"
#sourcecode[```rust
/// Port forwarder
#[derive(argh::FromArgs)]
struct Args {
    #[argh(positional)]
    listen_addr: SocketAddr,
    #[argh(positional)]
    connect_addr: SocketAddr,
}

let Args { listen_addr, connect_addr } 
            = argh::from_env();
```]
]

#centred-slide[
  = argh results

  #v60

   #table(
    columns: 5,
    [*target*], [*base size*], [*base reduction*], [*lean size*], [*lean reduction*],
    [amd64 linux],   [556k],  [#r(556,768)], [192k ], [#r(192,336)],
    [i686 musl],     [648k],  [#r(648,840)], [296k ], [#r(296,428)],
    [arm64 android], [528k],  [#r(528,708)], [228k ], [#r(228,360)],
    [windows64],     [480k],  [#r(480,680)], [216k ], [#r(216,368)],
    [arm64 mac],     [524k],  [#r(524,700)], [184k ], [#r(184,312)],
  )
]

#centred-slide[
  = Limiting Tokio's features

  #v20

  `cargo add tokio -F net,rt,macros,io-util`

   #table(
    columns: 5,
    [*target*], [*base size*], [*base reduction*], [*lean size*], [*lean reduction*],
    [amd64 linux],   [428k],  [#r(428,556)], [124k ], [#r(124,192)],
    [i686 musl],     [520k],  [#r(520,648)], [220k ], [#r(220,296)],
    [arm64 android], [404k],  [#r(404,528)], [156k ], [#r(156,228)],
    [windows64],     [400k],  [#r(400,480)], [192k ], [#r(192,216)],
    [arm64 mac],     [392k],  [#r(392,524)], [152k ], [#r(152,184)],
  )
]

#centred-slide[
  = Removing `anyhow`

  #v20

  `type Error = Box<dyn std::error::Error + Send + Sync>;`

  Errors still visible for CLI user, though not as nice.

   #table(
    columns: 5,
    [*target*], [*base size*], [*base reduction*], [*lean size*], [*lean reduction*],
    [amd64 linux],   [416k],  [#r(416,428)], [88k  ], [#r(88 ,124)],
    [i686 musl],     [508k],  [#r(508,520)], [100k ], [#r(100,220)],
    [arm64 android], [392k],  [#r(392,404)], [96k  ], [#r(96 ,156)],
    [windows64],     [388k],  [#r(388,400)], [152k ], [#r(152,192)],
    [arm64 mac],     [392k],  [#r(392,392)], [120k ], [#r(120,152)],
  )
]

#centred-slide[
  = What remains?

  #v40

```
2.0%  31.1% 18.6KiB base (our crate)
1.7%  26.1% 15.6KiB tokio
1.3%  19.5% 11.7KiB core
1.0%  15.6%  9.3KiB std
0.2%   2.7%  1.6KiB alloc
0.2%   2.6%  1.6KiB [Unknown]
0.1%   1.2%    714B argh
```
]
#centred-slide[
  = What remains?

  #v30

```
1.1%  16.1%  9.6KiB      base base::main::{{closure}}
1.0%  15.1%  9.0KiB      base base::main
0.3%   4.2%  2.5KiB     tokio tokio::...::poll
0.2%   2.6%  1.6KiB      core <Debug>::fmt
0.1%   2.2%  1.3KiB [Unknown] main
0.1%   1.8%  1.1KiB     tokio t.::r.::sch.::cu._th.::sh.2
0.1%   1.6%    983B     tokio copy_bidirectional::tr...on
0.1%   1.6%    978B      core core::fmt::Formatter::pad
```
]

#list-slide([What remains?])[

  #v20 

  - Actual implementation
  - Formatting

  #align(center)[#s[Apart from *UPX* (that would get us to 48k), there seems #linebreak() to be no easy ways to shrink this further.]]

  Note that we are *93% down* from the original size.
]


#centred-slide[
  = Part 2B: Shrinking challenge

  = (hacks)
]


#centred-slide[
  = Unportable trick: no_main

  We have *lost Windows* portability.
  
  There are still some error reports and CLI `--help` message.
  
  #s[There is ugly *startup function*, but CLI args and the algorithms niceties are *intact*. If stable is needed, the app can be turned into a staticlib and be used from a C wrapper.]

#v20 

  ```rust
#![no_main]
#[no_mangle]
fn main(mut argc: c_int, argv: *mut*mut c_char,
        _envp: *mut*mut c_char) -> c_int { ...
  ```
]

#centred-slide[
  = no_main results

  #v20

   #table(
    columns: 5,
    [*target*], [*base size*], [*base reduction*], [*lean size*], [*lean reduction*],
    [amd64 linux],   [396k],  [#r(396,416)], [76k  ], [#r(76 ,88 )],
    [i686 musl],     [484k],  [#r(484,508)], [88k  ], [#r(88 ,100)],
    [arm64 android], [376k],  [#r(376,392)], [84k  ], [#r(84 ,96 )],
    [arm64 mac],     [376k],  [#r(376,392)], [100k ], [#r(100,120)],
  )

  = Honourable mention: eyra

  Fully Rust C library, including startup object files.
]

#centred-slide[
  = Worse UX: only basic output

  Though we got rid of the formatted output, formatting code is *still present* in the executable.

   #table(
    columns: 5,
    [*target*], [*base size*], [*base reduction*], [*lean size*], [*lean reduction*],
    [amd64 linux],   [388k],  [#r(388,396)], [72k  ], [#r(72, 76 )],
    [i686 musl],     [480k],  [#r(480,484)], [84k  ], [#r(84, 88 )],
    [arm64 android], [368k],  [#r(368,376)], [80k  ], [#r(80, 84 )],
    [arm64 mac],     [360k],  [#r(360,376)], [100k ], [#r(100,100)],
  )
]

#centred-slide[
  = Investigating why fmt

Start by *ensuring* empty program is really *empty*.

  #sourcecode[```rust
fn main(...) -> c_int {
    return 0;
    ...
}
  ```]

```
File  .text Size     Crate Name
0.2%  14.3%  34B [Unknown] _start
0.0%   1.3%   3B [Unknown] main
```

Seems to be *close enough* to be considered *empty*.
]


#list-slide([Investigating why fmt])[
  #v20
  - As `return 0` is *moved lower* in `fn main`, we see functions gradually *appear* in `cargo bloat`.
  - *Formatting code appears* on the `rt.block_on(run(x))` line where we jump to our function.
  - So we begin with `return Ok(())` in *`fn run`* and *move it down* the similar way.
  - Formatting code *appears* when we get to the `TcpListener::bind(listen_addr).await`*`?`* line.
  - Specifically it's the *question mark* that causes it to be included.
]

#centred-slide[
  = Victory over fmt?

  #v40

  #sourcecode[```rust
struct DummyError;
impl<T:std::error::Error> From<T> for DummyError {
    fn from(_value: T) -> Self {
        DummyError
    }
}

async fn run(args: Args) -> Result<(), DummyError>;
  ```]
]


#centred-slide[
  = Victory over fmt?

  #v60

#table(
    columns: 5,
    [*target*], [*base size*], [*base reduction*], [*lean size*], [*lean reduction*],
    [amd64 linux],   [384k],  [#r(384,388)], [56k  ], [#r(56, 72 )],
    [i686 musl],     [480k],  [#r(480,480)], [76k  ], [#r(76, 84 )],
    [arm64 android], [364k],  [#r(364,368)], [60k  ], [#r(60, 80 )],
    [arm64 mac],     [360k],  [#r(360,360)], [84k  ], [#r(84, 100)],
  )
]


#centred-slide[
  = Victory over fmt?

  But the *spectre of `fmt`* is still hauting the project.

#show regex("fmt"): set text(fill: red)

  ```
 9.3KiB            main
 4.9KiB     tcpfwd tcpfwd::run::{{closure}}
 2.4KiB     tokio tokio::runtime::task::raw::poll
 1.6KiB     alloc <... core::fmt::Debug>::fmt
 1.1KiB     tokio tokio::r...::shutdown2
 1.0KiB     alloc <... core::fmt::Debug>::fmt
   983B     tokio copy_bidirectional::transfer_one_direction
   846B      core core::net::parser::<SocketAddr>::from_str
   665B     tokio tokio::...::scheduled_io::ScheduledIo::wake
   653B     tokio tokio::runtime::io::driver::Driver::turn
  ```
]

#centred-slide[
  = Extinguishing the last remnant of fmt

  #v40

  The culprit is *`io::Error::new`* calls in Tokio.

  Let's *patch* it away.

  `cargo override --path ~/src/tokio/tokio`

  #v20

  #s[Starting with `panic!()` at the top of `TcpListener::bind` and going down (i.e. *reordering the lines* for the `panic` to be lower and lower). Also just searching for `Error::new`.]
]


#centred-slide[
  = Extinguishing the last remnant of fmt

```diff
-        None => Err(io::Error::new(
-            io::ErrorKind::InvalidInput,
-            "invalid address family (not IPv4 or IPv6)",
-        )),               # (and 6 similar chunks)
+        None => Err(io::ErrorKind::InvalidInput.into()),
```

  #table(
    columns: 5,
    [*target*], [*base size*], [*base reduction*], [*lean size*], [*lean reduction*],
    [amd64 linux],   [384k],  [#r(384,384)], [48k  ], [#r(48, 56 )],
    [i686 musl],     [472k],  [#r(472,480)], [68k  ], [#r(68, 76 )],
    [arm64 android], [364k],  [#r(364,364)], [56k  ], [#r(56, 60 )],
    [arm64 mac],     [360k],  [#r(360,360)], [84k  ], [#r(84, 84 )],
  )
]

#centred-slide[
  = Manual implementation: no tokio

  ```
  cargo rm tokio
  cargo add mio -F net,os-poll
  ```

  Avoiding allocations, using my own `Option` to force `.bss`. #linebreak() *About 200 LOC*.

  #table(
    columns: 5,
    [*target*], [*base size*], [*base reduction*], [*lean size*], [*lean reduction*],
    [amd64 linux],   [296k],  [#r(296,384)], [12k  ], [#r(12, 48 )],
    [i686 musl],     [396k],  [#r(396,472)], [28k  ], [#r(28, 68 )],
    [arm64 android], [276k],  [#r(276,364)], [12k  ], [#r(12, 56 )],
    [arm64 mac],     [280k],  [#r(280,360)], [52k  ], [#r(52, 84 )],
  )
]


#centred-slide[
  = Outro
]


#centred-slide[

  = Helpful link

  #v60

  `https://github.com/johnthagen/min-sized-rust`

  #v40

  #qr-code("https://github.com/johnthagen/min-sized-rust", width: 6cm)

]


#list-slide([Shrinking methods review (easy)])[

  #v20

  Re-stating what it takes to shrink an executable.

  - Compilation *mode*
  - *Dependency* trimming
    - features
    - crate replacements
  - Shrinking the *requirements*
]


#list-slide([Shrinking methods review (hard)])[

  #v20

  Re-stating what it takes to shrink an executable.

  - Nightly-only *hacks*
  - `no_std` / `no_main` / `eyra` / `staticlib` / etc
  - Dependency trimming
    - *patching*
      - libstd patching
  - Manual adjustments to *inlining*, etc.
]


#list-slide("What typical Rust executable contains (1)")[
  #v60
  - startup code like `crt1.o`
  - C standard library (or symbols for linking to it)
  - Compiler functions library like `libgcc`
  - Rust's main function wrapper
  ...
]


#list-slide("What typical Rust executable contains (2)")[
  #v60
  ...
  - Rust standard library
      - Formatting code
      - Unwinder
      - Threading / TLS initializer
  - Your project's dependencies
  - Your own code
]



#centred-slide[
  = The end
]
